from pytest_alembic.config import Config
from pytest_alembic.plugin.fixtures import create_alembic_fixture
from pytest_alembic.runner import MigrationContext, runner

__all__ = [
    "Config",
    "create_alembic_fixture",
    "MigrationContext",
    "runner",
]



from pytest_alembic.plugin.plugin import PytestAlembicPlugin, TestOptionResolver


def pytest_addoption(parser):
    default_collector = TestOptionResolver.collect_test_definitions(
        default=True, experimental=False
    )
    default_tests = ", ".join(t.name for t in default_collector.available_tests.values())

    experimental_collector = TestOptionResolver.collect_test_definitions(
        default=False, experimental=True
    )
    experimental_tests = ", ".join(t.name for t in experimental_collector.available_tests.values())

    parser.addini(
        "pytest_alembic_include",
        "List of built-in tests to include. If specified, 'pytest_alembic_exclude' is ignored. "
        f"If both are omitted, all tests are included. Valid options include: {default_tests}",
    )
    parser.addini(
        "pytest_alembic_exclude",
        "List of built-in tests to exclude. Ignored if 'pytest_alembic_include' is specified."
        f"Valid options include: {default_tests}",
    )
    parser.addini(
        "pytest_alembic_include_experimental",
        "List of built-in experimental tests to include. Experimental tests must be explicitly "
        f"included. Valid options include: {experimental_tests}",
    )
    parser.addini(
        "pytest_alembic_tests_folder",
        "The location under which the built-in tests will be bound. This defaults to 'tests/' "
        "(the tests themselves then being executed from tests/pytest_alembic/*), the typical test "
        "location. However this can be customized if pytest is, for example, invoked from a parent "
        "directory like `pytest folder/tests`, or the tests are otherwise located at a different "
        "location, relative to `pytest`s invocation.",
        default="tests",
    )

    group = parser.getgroup("collect")
    group.addoption(
        "--test-alembic",
        action="store_true",
        default=False,
        help="Enable pytest-alembic built-in tests",
        dest="pytest_alembic_enabled",
    )
    group.addoption(
        "--alembic-exclude",
        default=None,
        help=f"List of built-in tests to exclude. Valid options include: {default_tests}",
        dest="pytest_alembic_exclude",
    )


def pytest_configure(config):
    config.addinivalue_line("markers", "alembic: Tests which use pytest-alembic.")


def pytest_sessionstart(session):
    if session.config.option.pytest_alembic_enabled:
        plugin = PytestAlembicPlugin(session.config)
        session.config.pluginmanager.register(plugin, "pytest-alembic")



import dataclasses
import warnings

import alembic.migration
from alembic.autogenerate import produce_migrations, render_python_code
from sqlalchemy import MetaData

from pytest_alembic.config import duplicate_alembic_config
from pytest_alembic.plugin.error import AlembicTestFailure
from pytest_alembic.runner import MigrationContext, run_connection_task
from pytest_alembic.tests.default import NOT_IMPLEMENTED_WARNING


def test_downgrade_leaves_no_trace(alembic_runner: MigrationContext):
    """Assert equal states of the MetaData before and after an upgrade/downgrade cycle.

    This test works by attempting to produce two autogenerated migrations.

    1. The first is the comparison between the original state of the database before the
       given migration's upgrade occurs, and the `MetaData` produced by having performed
       the upgrade.

       This should approximate the autogenerated migration that alembic
       would have generated to produce your upgraded database state itself.

    2. The 2nd is the comparison between the state of the database after having
       performed the upgrade -> downgrade cycle for this revision, and the same
       `MetaData` used in the first comparison.

       This should approximate what alembic would have autogenerated if you
       **actual** performed the downgrade on your database.

    In the event these two autogenerations do not match, it implies that your
    upgrade -> downgrade cycle produces a database state which is different
    (enough for alembic to detect) from the state of the database without having
    performed the migration at all.

    **note** this isn't perfect! Alembic autogeneration will not detect many
    kinds of changes! If you encounter some scenario in which this does not
    detect a change you'd expect it to, alembic already has extensive ability
    to customize and extend the autogeneration capabilities.
    """
    run_connection_task(
        alembic_runner.connection, _test_downgrade_leaves_no_trace, alembic_runner=alembic_runner
    )


def _test_downgrade_leaves_no_trace(connection, alembic_runner: MigrationContext):
    # Swap the original engine for a connection to enable us to rollback the transaction
    # midway through.
    alembic_config = duplicate_alembic_config(alembic_runner.command_executor.alembic_config)
    alembic_config.attributes["connection"] = connection

    alembic_runner = dataclasses.replace(
        alembic_runner,
        connection=connection,
        command_executor=dataclasses.replace(
            alembic_runner.command_executor, alembic_config=alembic_config
        ),
    )

    history = alembic_runner.history
    revisions = history.revisions[:-1]
    if len(revisions) == 1:
        return

    below_minimum = alembic_runner.config.minimum_downgrade_revision is not None
    for revision in revisions:
        if below_minimum and revision == alembic_runner.config.minimum_downgrade_revision:
            below_minimum = False

        # Semantically, we'll solely upgrade for as long as we're below the `minimum_downgrade_revision`,
        # if set. If not set, then this is always done.
        if not below_minimum:
            # Leaves the database in its previous state, to avoid subtle upgrade -> downgrade issues.
            check_revision_cycle(alembic_runner, connection, revision)

        # So we need to proceed by one.
        alembic_runner.migrate_up_to(revision)

        if hasattr(connection, "commit"):
            connection.commit()


def check_revision_cycle(alembic_runner, connection, original_revision):
    migration_context = alembic.migration.MigrationContext.configure(connection)

    # We first need to produce a `MetaData` which represents the state of the database
    # we're trying to get to.
    with connection.begin() as trans:
        alembic_runner.migrate_up_one()
        upgrade_revision = alembic_runner.current

        upgrade_metadata = MetaData()
        upgrade_metadata.reflect(connection)

        # Having procured the target `MetaData`, we need the database back in its original state.
        trans.rollback()

    with connection.begin() as trans:
        # Produce a canonically autogenerated upgrade relative to the original.
        autogenerated_upgrade = produce_migrations(migration_context, upgrade_metadata)
        rendered_autogenerated_upgrade = render_python_code(autogenerated_upgrade.upgrade_ops)

        # Now, we can perform the upgrade -> downgrade cycle!
        alembic_runner.migrate_up_one()
        try:
            alembic_runner.migrate_down_one()
        except NotImplementedError:
            # In the event of a `NotImplementedError`, we should have the same semantics,
            # as-if `minimum_downgrade_revision` was specified, but we'll emit a warning
            # to suggest that feature is used instead.
            warnings.warn(NOT_IMPLEMENTED_WARNING.format(revision=upgrade_revision))

        else:
            downgrade_metadata = MetaData()
            downgrade_metadata.reflect(connection)

            # Produce a canonically autogenerated upgrade relative to the post-downgrade state.
            autogenerated_post_downgrade = produce_migrations(migration_context, upgrade_metadata)
            rendered_autogenerated_post_downgrade = render_python_code(
                autogenerated_post_downgrade.upgrade_ops
            )

            if rendered_autogenerated_upgrade != rendered_autogenerated_post_downgrade:
                raise AlembicTestFailure(
                    (
                        f"There is a difference between the pre-'{upgrade_revision}'-upgrade `MetaData`, "
                        f"and the post-'{upgrade_revision}'-downgrade `MetaData`. This implies that the "
                        "upgrade performs some set of DDL changes which the downgrade does not "
                        "precisely undo."
                    ),
                    context=[
                        (
                            f"DDL diff for {original_revision} -> {upgrade_revision}",
                            rendered_autogenerated_upgrade,
                        ),
                        (
                            f"DDL diff after performing the {upgrade_revision} -> {original_revision} downgrade",
                            rendered_autogenerated_post_downgrade,
                        ),
                    ],
                )
        finally:
            # **This** rollback is to ensure we leave the database back in it's original state for the next revision.
            trans.rollback()



import collections
import itertools
from dataclasses import dataclass
from typing import Dict, List, Optional, Tuple

from alembic.script.revision import RevisionMap


@dataclass
class AlembicHistory:
    map: RevisionMap
    revisions: List[str]
    revision_indices: Dict[str, int]
    revisions_by_index: Dict[int, str]

    @classmethod
    def parse(cls, revision_map: RevisionMap) -> "AlembicHistory":
        """Extract the set of migration revision hashes from alembic's notion of the history."""
        revision_hashes = ["heads"]

        history = revision_map.iterate_revisions("heads", "base")
        for script in history:
            revision = script.revision
            revision_hashes.append(revision)
        revision_hashes.append("base")

        revisions = list(reversed(revision_hashes))
        revision_indices = {revision: i for i, revision in enumerate(revisions)}
        revisions_by_index = {v: k for k, v in revision_indices.items()}
        return cls(
            map=revision_map,
            revisions=revisions,
            revision_indices=revision_indices,
            revisions_by_index=revisions_by_index,
        )

    def validate_revision(self, revision):
        # Given that 'heads' seems to be strictly more powerful, coerce singular 'head'
        # to 'heads'.
        if revision == "head":
            revision = "heads"

        if revision not in self.revision_indices:
            raise ValueError(f"Revision {revision} is not a valid revision in alembic's history")
        return revision

    def previous_revision(self, revision: str) -> Optional[str]:
        revision = self.validate_revision(revision)
        revision_index = self.revision_indices[revision]
        return self.revisions_by_index.get(revision_index - 1)

    def next_revision(self, revision: str) -> Optional[str]:
        revision = self.validate_revision(revision)
        revision_index = self.revision_indices[revision]
        return self.revisions_by_index.get(revision_index + 1)

    def revision_range(self, current_revision: str, dest_revision: str) -> List[str]:
        current_revision = self.validate_revision(current_revision)
        dest_revision = self.validate_revision(dest_revision)
        start_index = self.revision_indices[current_revision]
        end_index = self.revision_indices[dest_revision]
        return [self.revisions[index] for index in range(start_index, end_index + 1)]

    def revision_window(self, current_revision: str, dest_revision: str) -> List[Tuple[str, str]]:
        revision_range = self.revision_range(current_revision, dest_revision)
        return list(
            zip(  # type: ignore
                *(
                    collections.deque(itertools.islice(it, i), 0) or it
                    for i, it in enumerate(itertools.tee(revision_range, 2))
                )
            )
        )



import contextlib
import io
from dataclasses import dataclass, field
from io import StringIO
from typing import Dict, List, Optional, Union

import alembic
import alembic.config
from alembic.runtime.environment import EnvironmentContext
from alembic.script.base import ScriptDirectory
from sqlalchemy import MetaData, Table
from sqlalchemy.engine import Connection

from pytest_alembic.config import Config


@dataclass
class CommandExecutor:
    alembic_config: alembic.config.Config
    stdout: StringIO
    stream_position: int
    script: ScriptDirectory

    @classmethod
    def from_config(cls, config: Config):
        stdout = StringIO()
        alembic_config = config.make_alembic_config(stdout)
        return cls(
            alembic_config=alembic_config,
            stdout=stdout,
            stream_position=0,
            script=ScriptDirectory.from_config(alembic_config),
        )

    def configure(self, **kwargs):
        for key, value in kwargs.items():
            self.alembic_config.attributes[key] = value

    def run_command(self, command, *args, **kwargs):
        self.stream_position = self.stdout.tell()

        executable_command = getattr(alembic.command, command)
        try:
            # Hide the (relatively) worthless logs of the upgrade revision path, it just clogs
            # up the logs when errors actually occur, but without providing any context.
            buffer = io.StringIO()
            with contextlib.redirect_stderr(buffer):
                executable_command(self.alembic_config, *args, **kwargs)
        except alembic.util.exc.CommandError as e:
            raise RuntimeError(e)

        self.stdout.seek(self.stream_position)
        return self.stdout.readlines()

    def heads(self):
        return [rev.revision for rev in self.script.get_revisions("heads")]

    def upgrade(self, revision):
        """Upgrade to the given `revision`."""

        def upgrade(rev, _):
            return self.script._upgrade_revs(revision, rev)

        self._run_env(upgrade, revision)

    def downgrade(self, revision):
        """Downgrade to the given `revision`."""

        def downgrade(rev, _):
            return self.script._downgrade_revs(revision, rev)

        self._run_env(downgrade, revision)

    def _run_env(self, fn, revision=None):
        """Execute the migrations' env.py, given some function to execute."""
        dont_mutate = revision is None
        with EnvironmentContext(
            self.alembic_config,
            self.script,
            fn=fn,
            destination_rev=revision,
            dont_mutate=dont_mutate,
        ):
            self.script.run_env()


@dataclass
class ConnectionExecutor:
    metadatas: Dict[str, MetaData] = field(default_factory=dict)

    def metadata(self, revision: str) -> MetaData:
        metadata = self.metadatas.get(revision)
        if metadata is None:
            metadata = MetaData()
            self.metadatas[revision] = metadata

        return metadata

    def table(self, connection, revision: str, name: str, schema: Optional[str] = None) -> Table:
        meta = self.metadata(revision)
        if name in meta.tables:
            return meta.tables[name]

        return Table(name, meta, schema=schema, autoload_with=connection)

    def table_insert(
        self,
        connection: Connection,
        revision: str,
        data: Union[Dict, List],
        tablename: Optional[str] = None,
        schema: Optional[str] = None,
    ):
        if isinstance(data, dict):
            data = [data]

        for item in data:
            _tablename = item.get("__tablename__", None)
            table = _tablename or tablename

            if table is None:
                raise ValueError(
                    "No table name provided as either `table` argument, or '__tablename__' key in `data`."
                )

            try:
                # Attempt to parse the schema out of the tablename
                schema, table = table.split(".", 1)
            except ValueError:
                # However, if it doesn't work, both `table` and `schema` are in scope, so failure is fine.
                pass

            table = self.table(connection, revision, table, schema=schema)
            values = {k: v for k, v in item.items() if k != "__tablename__"}
            connection.execute(table.insert().values(values))



"""Collect module and table information from the env.py context.

This module is executed within a subprocess in order to maintain a clean
python import state. Thus the `print` statements are the mechanism through
which this script communicates the environment state back to the
parent process.
"""
import gc
import json
import os
import sys

from alembic.config import Config
from alembic.runtime.environment import EnvironmentContext
from alembic.script import ScriptDirectory

try:
    from sqlalchemy.ext.declarative import DeclarativeMeta
except ImportError:  # pragma: no cover
    from sqlalchemy.declarative import DeclarativeMeta


def run():  # pragma: no cover
    """Execute alembic in order to spin up the alembic environment.

    Cannot be covered with coverage due to the subprocess environment in which it's
    executed.
    """
    _, url, offline_str, async_str = sys.argv
    offline = bool(int(offline_str))
    async_ = bool(int(async_str))

    config = Config("alembic.ini")
    config.set_section_option(config.config_ini_section, "sqlalchemy.url", url)

    try:
        config.attributes["connection"] = create_connectable(url, async_)
    except Exception:  # nosec
        pass

    script = ScriptDirectory.from_config(config)

    with EnvironmentContext(config, script, fn=environment_context_fn, as_sql=offline):
        script.run_env()


def create_connectable(url, async_=False):
    if async_:
        from sqlalchemy.ext.asyncio import create_async_engine

        connectable = create_async_engine(url)
    else:
        from sqlalchemy import create_engine

        connectable = create_engine(url).connect()
    return connectable


def environment_context_fn(_, migration_context):
    """Print out the alembic environment context (given an alembic migration context)."""
    target_metadata = migration_context.opts["target_metadata"]
    if target_metadata:
        tables = list(target_metadata.tables)
        modules = list(identify_modules(target_metadata))
    else:
        tables = []
        modules = []

    result = {
        "modules": modules,
        "tables": tables,
    }

    # Remember, given that we're in a subprocess, `print` is our output mechanism.
    output = json.dumps(result)
    print(f"<pytest-alembic>{output}</pytest-alembic>")

    return []


def identify_modules(target_metadata):
    """Find all model referrers of the alembic `target_metadata`."""
    for referrer in gc.get_referrers(target_metadata):
        dict_referrer = isinstance(referrer, dict)
        if not dict_referrer:
            continue

        model_base = get_model_base(referrer, target_metadata)
        if model_base:
            for referrer in gc.get_referrers(model_base):
                yield from get_referrer_module(referrer)

        yield from get_referrer_module(referrer)


def get_model_base(referrer, target_metadata):
    """Find all model referrers of any model base referents of the `target_metadata`."""
    metadata = referrer.get("metadata")
    if not metadata or metadata is not target_metadata:
        return

    for referrer in gc.get_referrers(referrer):
        if isinstance(referrer, DeclarativeMeta):
            return referrer


def get_referrer_module(referrer):
    """Distinguish actual candidate modules from those which cannot be modules.

    Namely: non-modules, alembic-related things, scripts and such.
    """
    dict_referrer = isinstance(referrer, dict)

    loader = dict_referrer and referrer.get("__loader__")
    loader_name = getattr(loader, "name", None)

    if loader_name in {None, "__main__", "env_py"}:
        return

    # If we've already ruled out the main and/or env.py,
    # then this is likely the module we want.
    yield loader_name


if __name__ == "__main__":
    if os.environ.get("COVERAGE_PROCESS_START"):
        import coverage

        coverage.process_startup()

    run()



import textwrap

from _pytest._code.code import FormattedExcinfo


class AlembicTestFailure(AssertionError):
    def __init__(self, message, context=None):
        super().__init__(message)
        self.context = context


class AlembicReprError:
    def __init__(self, exce, item):
        self.exce = exce
        self.item = item

    def toterminal(self, tw):
        """Print out a custom error message to the terminal."""
        exc = self.exce.value
        context = exc.context

        if context:
            for title, item in context:
                tw.line(title + ":", white=True, bold=True)
                tw.line(textwrap.indent(item, "    "), red=True)
                tw.line("")

        e = FormattedExcinfo()
        lines = e.get_exconly(self.exce)

        tw.line("Errors:", white=True, bold=True)
        for line in lines:
            tw.line(line, red=True, bold=True)



from dataclasses import dataclass, field
from typing import Any, Dict, Optional, Union

import alembic.config


@dataclass
class Config:
    """Pytest-alembic configuration options.

    - `config_options`: Meant to simplify the creation of ``alembic.config.Config``
       objects. Supply keys common to customization in alembic configuration. For
       example:

       - file/config_file_name (commonly alembic.ini)
       - script_location
       - sqlalchemy.url
       - target_metadata
       - process_revision_directives
       - include_schemas

    - Both `before_revision_data` and `at_revision_data` are described in detail
      in :ref:`Custom data`.

    - :code:`minimum_downgrade_revision` can be used to set a lower bound on the
      **downgrade** migrations which are run built-in tests like ``test_up_down_consistency``
      and ``test_downgrade_leaves_no_trace``.

    For example:
        >>> import pytest

        >>> @pytest.fixture
        ... def alembic_config():
        ...    return Config(minimum_downgrade_revision='abcde12345')

        This would essentially short-circuit and avoid running the downgrade
        migrations **including and below** this migration.

    .. note::

       If a downgrade raises a ``NotImplementedError``, it will have the same effect
       as a ``minimum_downgrade_revision``, but will emit a warning suggesting
       the use of this feature instead.
    """

    config_options: Dict[str, Any] = field(default_factory=dict)
    alembic_config: Optional[alembic.config.Config] = None

    before_revision_data: Optional[Union[Dict, "RevisionSpec"]] = None
    at_revision_data: Optional[Union[Dict, "RevisionSpec"]] = None

    minimum_downgrade_revision: Optional[str] = None

    @classmethod
    def from_raw_config(
        cls, raw_config: Union[Dict[str, Any], alembic.config.Config, "Config", None] = None
    ):
        """Adapt between pre-produced alembic config and raw config options.

        Allows one to specify raw pytest-alembic config options through raw dictionary,
        as well as being flexible enough to allow a literal alembic Config object.

        Examples:
            >>> Config.from_raw_config()
            Config(config_options={}, alembic_config=None, before_revision_data=None, at_revision_data=None, minimum_downgrade_revision=None)

            >>> Config.from_raw_config({'minimum_downgrade_revision': 'abc123'})
            Config(config_options={}, alembic_config=None, before_revision_data=None, at_revision_data=None, minimum_downgrade_revision='abc123')

            >>> Config.from_raw_config(Config(minimum_downgrade_revision='abc123'))
            Config(config_options={}, alembic_config=None, before_revision_data=None, at_revision_data=None, minimum_downgrade_revision='abc123')
        """
        if raw_config is None:
            return cls()

        if isinstance(raw_config, alembic.config.Config):
            return cls(alembic_config=raw_config)

        if isinstance(raw_config, Config):
            return raw_config

        before_data = raw_config.pop("before_revision_data", None)
        at_data = raw_config.pop("at_revision_data", None)
        minimum_downgrade_revision = raw_config.pop("minimum_downgrade_revision", None)
        return cls(
            config_options=raw_config,
            alembic_config=None,
            before_revision_data=before_data,
            at_revision_data=at_data,
            minimum_downgrade_revision=minimum_downgrade_revision,
        )

    def make_alembic_config(self, stdout):
        file = (
            self.config_options.get("file")
            or self.config_options.get("config_file_name")
            or "alembic.ini"
        )

        alembic_config = self.config_options.get("alembic_config")

        if not alembic_config and self.alembic_config:
            alembic_config = self.alembic_config
            alembic_config.stdout = stdout
        else:
            alembic_config = alembic.config.Config(file, stdout=stdout)

        sqlalchemy_url = self.config_options.get("sqlalchemy.url")
        if sqlalchemy_url:
            alembic_config.set_main_option("sqlalchemy.url", sqlalchemy_url)

        # Only set script_location if set.
        script_location = self.config_options.get("script_location")
        if script_location:
            alembic_config.set_section_option("alembic", "script_location", script_location)
        elif not alembic_config.get_section_option("alembic", "script_location"):
            # Or in the event that it's not set after already having loaded the config.
            alembic_config.set_main_option("script_location", "migrations")

        target_metadata = self.config_options.get("target_metadata")
        alembic_config.attributes["target_metadata"] = target_metadata

        process_revision_directives = self.config_options.get("process_revision_directives")
        alembic_config.attributes["process_revision_directives"] = process_revision_directives

        include_schemas = self.config_options.get("include_schemas", True)
        alembic_config.attributes["include_schemas"] = include_schemas

        return alembic_config


def duplicate_alembic_config(config: alembic.config.Config):
    return alembic.config.Config(
        config.config_file_name,
        ini_section=config.config_ini_section,
        output_buffer=config.output_buffer,
        stdout=config.stdout,
        cmd_opts=config.cmd_opts,
        config_args=config.config_args,
        attributes=config.attributes,
    )


# isort: split
from pytest_alembic.revision_data import RevisionSpec  # noqa



from dataclasses import dataclass
from typing import Dict, List, Union


@dataclass
class RevisionSpec:
    """Describe a set of valid database data at a set of revisions."""

    data: Dict[str, Union[Dict, List[Dict]]]

    @classmethod
    def parse(cls, data: Union[None, "RevisionSpec", Dict[str, Union[Dict, List[Dict]]]]):
        """Parse a raw dict structure into a `RevisionSpec`."""
        if not data:
            return cls({})

        if isinstance(data, RevisionSpec):
            return data

        return cls(data)

    def get(self, revision: str) -> Union[Dict, List[Dict]]:
        """Get the database data described at a particular revision."""
        return self.data.get(revision, [])


@dataclass
class RevisionData:
    """Describe the data which should exist at given revisions when performing upgrades."""

    before_revision_data: RevisionSpec
    at_revision_data: RevisionSpec

    @classmethod
    def from_config(cls, config: "Config"):
        """Produce a `RevisionData` from raw configuration from :func:`alembic_config`."""
        return cls(
            before_revision_data=RevisionSpec.parse(config.before_revision_data),
            at_revision_data=RevisionSpec.parse(config.at_revision_data),
        )

    def get(self, revision: str, revision_data: Union[Dict, List[Dict]]):
        if isinstance(revision_data, Dict):
            yield revision_data
        else:
            for item in revision_data:
                yield item

    def get_before(self, revision: str) -> Union[Dict, List[Dict]]:
        """Yield the individual data insertions which should occur before the given revision."""
        before_revision_data = self.before_revision_data.get(revision)
        return self.get(revision, before_revision_data)

    def get_at(self, revision: str) -> Union[Dict, List[Dict]]:
        """Yield the individual data insertions which should occur upon reaching the given revision."""
        at_revision_data = self.at_revision_data.get(revision)
        return self.get(revision, at_revision_data)


# isort: split
from pytest_alembic.config import Config  # noqa



import importlib
import json
import logging
import pkgutil
import re
import subprocess  # nosec
from typing import List, Optional, Set, Tuple

from sqlalchemy import MetaData

from pytest_alembic.plugin.error import AlembicTestFailure
from pytest_alembic.runner import MigrationContext

try:
    from sqlalchemy.ext.declarative import DeclarativeMeta
except ImportError:  # pragma: no cover
    from sqlalchemy.declarative import DeclarativeMeta

try:
    from sqlalchemy.ext.asyncio import AsyncEngine
except ImportError:
    AsyncEngine = None


log = logging.getLogger(__name__)


def test_all_models_register_on_metadata(
    alembic_runner: MigrationContext,
    model_package: Optional[str] = None,
    offline: bool = False,
    async_: bool = None,
):
    """Assert that all tables defined on your `MetaData`, are imported in the `env.py`.

    We'll call a "bare import", the minimum import of the model base/MetaData which is
    typically performed in the `env.py`. Generally this might look something like:

    >>>
    >> from foo.models import Base
    >> target_metadata = Base.metadata
    >> ...
    >> context.configure(..., target_metadata=target_metadata)

    You'll notice here, that if `foo.models` does not, itself, import a theoretical
    `foo.models.bar` module, then that model will fail to register on the metadata
    and as far as alembic's `--autogenerate` feature goes, it does not exist.

    Generally one might notice this as an `--autogenerate` run dropping tables for
    which you have table definitions, at which point you'll go back and make sure
    to import said model somewhere in the normal loading process for `foo.models`.

    Thus this test compares such an import against the opposite scenario, one where
    every single possible module has been imported, and we therefore "**know**" that
    all defined models/tables will be attached to the MetaData. If there's a
    difference in the set of tables on the MetaData, then we have an issue!

    **Note** This test is distinct from the `test_model_definitions_match_ddl`
    test because imports can have side-effects. Thus when normal tests run in the
    context of your test run, you may have already imported models from modules
    transitively somewhere else in tree of imports required to run your tests.

    Therefore this test goes to lengths to collect a clean python interpreter
    which **only** imports the portion of code which would have been imported by
    the `env.py` itself.

    Arguments:
        alembic_runner: The alembic runner instance used to run the test.
        model_package: An optional way to explicitly the package in which models
            reside, in the event that automatic detection fails.
        offline: Whether to run the internal migration step as offline. This can
            be useful, depending on how your env.py is set up, to avoid executing
            certain portions of your env.py, if they're incompatible with execution
            in the context of this test.
        async_: Whether to produce an async engine in the internal engine creation
            step. Defaults to `None` (i.e. automatic detection).
    """
    # If `async_` is None, then we should automatically detect whether to run in
    # async mode. If `AsyncEngine` is None, then we're not running on a version
    # of sqlalchemy which supports it.
    if (
        async_ is None
        and AsyncEngine is not None
        and isinstance(alembic_runner.connection, AsyncEngine)
    ):
        async_ = True

    modules, bare_tables = get_bare_import_tableset(
        str(alembic_runner.connection.url),
        async_=bool(async_),
        offline=offline,
    )
    if model_package:
        modules = [model_package]

    full_tables = get_full_tableset(*modules)

    if bare_tables != full_tables:
        diff = bare_tables.symmetric_difference(full_tables)
        diff_str = ", ".join(diff)
        raise AlembicTestFailure(
            "There was a difference detected between the set of tables registered on your "
            "`MetaData` during the course of loading your `env.py` compared with the set of "
            f"tables registered after importing all modules below the provided '{model_package}' "
            "module/package.\n\n"
            "This commonly happens if you are missing an import path from the root of your "
            "models/MetaData to some set of models defined elsewhere. Given that your `env.py` "
            "likely only directly imports your model base or MetaData, this can be a problem! "
            "As far as alembic is concerned, affected tables do not exist, and will likely "
            "be indicated by alembic to be dropped, should you attempt a `revision --autogenerate`"
            "command.\n\n"
            f"The following tables are affected: {diff_str}."
        )


def get_bare_import_tableset(
    url: str, offline: bool = False, async_: bool = False
) -> Tuple[List[str], Set[str]]:
    """Get the set of tables which would have been added to the metadata on a bare ma.models import.

    Importantly, this cannot simply import the MetaData directly, as that may have already
    had side-effects performed on it such that it's no longer representative of if one were to
    directly import the MetaData alone.

    Thus, we start a python subprocess which does the most minimal import we can perform to extract the MetaData,
    and return the set of tables attached to it.
    """

    command = [
        "python",
        "-m",
        "pytest_alembic.tests.experimental.collect_clean_alembic_environment",
        url,
        str(offline.real),
        str(async_.real),
    ]
    try:
        result = subprocess.run(  # nosec
            command,
            check=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            universal_newlines=True,
        )
    except subprocess.CalledProcessError as e:  # pragma: no cover
        raise AlembicTestFailure(e.stderr)

    parsed_result = parse_collection_output(result.stdout)
    tablenames = set(parsed_result["tables"])
    modules = sorted(parsed_result["modules"])
    return modules, tablenames


def parse_collection_output(raw_output: str):
    # The default env.py offline execution mode, emits extra output that is
    # irrelevant to the expected output that a script execution would normally
    # produce, so we surround it with sentinel content.
    match = re.match(r"<pytest-alembic>(.*)</pytest-alembic>", raw_output)
    if match:
        return json.loads(match.group(1))

    # Indicative of a bug in the implementation of the subprocess code!
    raise RuntimeError(raw_output)


def get_full_tableset(*module_names: str) -> Set[str]:
    """Get the set of full set of tables which are defined.

    The theory is that if we import every module in the tree below ma.models, we should
    cause the import side-effect of all models being attached to the model base.

    Once done, we should end up with the theoretical maximum of tables included in the MetaData.
    """
    metadata = None
    for module_name in module_names:
        modules = traverse_modules(module_name)
        try:
            module, *_ = modules
        except ValueError:
            raise AlembicTestFailure(f"Invalid module name: {module_name}")

        for item in module.__dict__.values():
            if isinstance(item, MetaData):
                metadata = item
                break
            if isinstance(item, DeclarativeMeta):
                metadata = item.metadata
                break

    if metadata is None:
        raise AlembicTestFailure(
            "Unable to locate a MetaData reference in the local context of your `env.py`. This "
            "test requires a local reference to either a `DeclarativeMeta` (i.e. declarative model "
            "base), or a `MetaData` to exist in your `env.py`."
        )

    tables = metadata.tables
    tablenames = set(tables.keys())
    return tablenames


def traverse_modules(
    package_name: str,
    import_module=importlib.import_module,
    walk_packages=pkgutil.walk_packages,
):
    """Dynamically traverse the tree of packages below a given root and import them.

    Note this will perform the import of each module, which is an operation which can
    cause side-effects!
    """

    try:
        package = import_module(package_name)
    except ImportError:
        return

    if package.__package__:
        if not hasattr(package, "__path__"):
            return

        yield package

        package_path = package.__path__
        for _, name, is_package in walk_packages(package_path):
            full_name = f"{package.__name__}.{name}"

            try:
                yield import_module(full_name)
            except ImportError:
                continue

            if is_package:
                yield from traverse_modules(full_name)

    else:
        name = package.__name__
        yield import_module(name)



from typing import Any, Dict, Union

import alembic.config
import pytest
import sqlalchemy

import pytest_alembic
from pytest_alembic.config import Config


def create_alembic_fixture(raw_config=None):
    """Create a new fixture `alembic_runner`-like fixture.

    In many cases, this function should not be strictly necessary. You **can**
    generally rely solely on the :code:`--test-alembic` flag, automatic insertion
    of tests, and the :func:`alembic_runner` fixture.

    However this may be useful in some situations:

    - If you would generally prefer to avoid the :code:`--test-alembic` flag and
      automatic test insertion, this is the function for you!
    - If you have multiple alembic histories and therefore require more than one
      fixture, you will **minimally** need to use this for the 2nd history (if
      not both)

    Examples:
        >>> from pytest_alembic import tests
        >>>
        >>> alembic = create_alembic_fixture()
        >>>
        >>> def test_upgrade_head(alembic):
        ...     tests.test_upgrade_head(alembic)
        >>>
        >>> def test_specific_migration(alembic):
        ...     alembic_runner.migrate_up_to('xxxxxxx')
        ...     assert ...

        Config can also be supplied similarly to the :func:`alembic_config` fixture.

        >>> alembic = create_alembic_fixture({'file': 'migrations.ini'})
    """

    @pytest.fixture
    def _(alembic_engine):
        config = Config.from_raw_config(raw_config)
        with pytest_alembic.runner(config=config, engine=alembic_engine) as runner:
            yield runner

    return _


@pytest.fixture
def alembic_runner(alembic_config, alembic_engine):
    """Produce the primary alembic migration context in which to execute alembic tests.

    This fixture allows authoring custom tests which are specific to your particular
    migration history.

    Examples:
        >>> def test_specific_migration(alembic_runner):
        ...     alembic_runner.migrate_up_to('xxxxxxx')
        ...     assert ...
    """
    config = Config.from_raw_config(alembic_config)
    with pytest_alembic.runner(config=config, engine=alembic_engine) as runner:
        yield runner


@pytest.fixture
def alembic_config() -> Union[Dict[str, Any], alembic.config.Config, Config]:
    """Override this fixture to configure the exact alembic context setup required.

    The return value of this fixture can be one of a few types.

    - If you're only configuring alembic-native configuration, a :class:`alembic.config.Config`
      object is accepted as configuration. This largely leaves pytest-alembic out
      of the setup, so depending on your settings, might be the way to go.

    - If you only have a couple of options to set, you might choose to return
      a ``Dict``.

      The following common alembic config options are accepted as keys.

        - file/config_file_name (commonly alembic.ini)
        - script_location
        - sqlalchemy.url
        - target_metadata
        - process_revision_directives
        - include_schemas

      Additionally you can send a `file` key (akin to `alembic -c`), should your
      `alembic.ini` be otherwise named.

      Note that values here, represent net-additive options on top of what you might
      already have configured in your `env.py`. You should generally prefer to
      configure your `env.py` however you like it and omit such options here.

      You may also use this dict to set pytest-alembic specific features:

        - before_revision_data
        - at_revision_data
        - minimum_downgrade_revision

    - You can also directly return a :ref:`Config` class instance.
      This is your only option if you want to use both pytest-alembic specific features
      **and** construct your own :class:`alembic.config.Config`.

    Examples:
        >>> @pytest.fixture
        ... def alembic_config():
        ...     return {'file': 'migrations.ini'}

        >>> @pytest.fixture
        ... def alembic_config():
        ...     alembic_config = alembic.config.Config()
        ...     alembic_config.set_main_option("script_location", ...)
        ...     return alembic_config
    """
    return {}


@pytest.fixture
def alembic_engine():
    """Override this fixture to provide pytest-alembic powered tests with a database handle."""
    return sqlalchemy.create_engine("sqlite:///")



from pytest_alembic.tests.experimental.all_models_register_on_metadata import (
    test_all_models_register_on_metadata,
)
from pytest_alembic.tests.experimental.downgrade_leaves_no_trace import (
    test_downgrade_leaves_no_trace,
)

__all__ = [
    "test_all_models_register_on_metadata",
    "test_downgrade_leaves_no_trace",
]



import logging
import warnings

import pytest
from alembic.autogenerate.api import AutogenContext
from alembic.autogenerate.render import _render_cmd_body

from pytest_alembic.plugin.error import AlembicTestFailure

log = logging.getLogger(__name__)


NOT_IMPLEMENTED_WARNING = (
    "The {revision} downgrade raised `NotImplementedError`, which short-circuited the downgrade "
    "operation and may have passed the test. If intended, and downgrades can not safely be performed "
    "below this migration, see 'minimum_downgrade_revision' configuration to avoid this warning."
)


@pytest.mark.alembic
def test_single_head_revision(alembic_runner):
    """Assert that there only exists one head revision.

    We're not sure what realistic scenario involves a diverging history to be desirable. We
    have only seen it be the result of uncaught merge conflicts resulting in a diverged history,
    which lazily breaks during deployment.
    """
    heads = alembic_runner.heads
    head_count = len(heads)

    if head_count != 1:
        heads = "\n".join([h.strip() for h in heads])
        raise AlembicTestFailure(
            "Expected 1 head revision, found {}".format(head_count),
            context=[("Heads", heads)],
        )


@pytest.mark.alembic
def test_upgrade(alembic_runner):
    """Assert that the revision history can be run through from base to head."""
    try:
        alembic_runner.migrate_up_to("heads")
    except RuntimeError as e:
        raise AlembicTestFailure(
            "Failed to upgrade to the head revision. This means the historical chain from an "
            "empty database, to the current revision is not possible.",
            context=[("Alembic Error", str(e))],
        )


@pytest.mark.alembic
def test_model_definitions_match_ddl(alembic_runner):
    """Assert that the state of the migrations matches the state of the models describing the DDL.

    In general, the set of migrations in the history should coalesce into DDL which is described
    by the current set of models. Therefore, a call to `revision --autogenerate` should always
    generate an empty migration (e.g. find no difference between your database (i.e. migrations
    history) and your models).
    """

    def verify_is_empty_revision(migration_context, __, directives):
        script = directives[0]

        migration_is_empty = script.upgrade_ops.is_empty()
        if not migration_is_empty:
            autogen_context = AutogenContext(migration_context)
            rendered_upgrade = _render_cmd_body(script.upgrade_ops, autogen_context)

            if not migration_is_empty:
                raise AlembicTestFailure(
                    "The models describing the DDL of your database are out of sync with the set of "
                    "steps described in the revision history. This usually means that someone has "
                    "made manual changes to the database's DDL, or some model has been changed "
                    "without also generating a migration to describe that change.",
                    context=[
                        (
                            "The upgrade which would have been generated would look like",
                            rendered_upgrade,
                        )
                    ],
                )

    test_upgrade(alembic_runner)
    alembic_runner.generate_revision(
        message="test revision",
        autogenerate=True,
        process_revision_directives=verify_is_empty_revision,
    )


@pytest.mark.alembic
def test_up_down_consistency(alembic_runner):
    """Assert that all downgrades succeed.

    While downgrading may not be lossless operation data-wise, there’s a theory of database
    migrations that says that the revisions in existence for a database should be able to go
    from an entirely blank schema to the finished product, and back again.

    Individually upgrade to ensure that it's clear which revision caused the failure.
    """
    for revision in alembic_runner.history.revisions:
        try:
            alembic_runner.migrate_up_to(revision)
        except Exception as e:
            raise AlembicTestFailure(
                "Failed to upgrade through each revision individually.",
                context=[("Failing Revision", revision), ("Alembic Error", str(e))],
            )

    # Skip the `heads` revision. Caused by new alembic warning in 1.6.x.
    down_revisions = list(reversed(alembic_runner.history.revisions[:-1]))

    index = 0
    for index, revision in enumerate(down_revisions):
        if alembic_runner.config.minimum_downgrade_revision == revision:
            # If there is a minimum_downgrade_revision, stop downgrading here.
            break

        try:
            alembic_runner.migrate_down_to(revision)
        except NotImplementedError:
            # In the event of a `NotImplementedError`, we should have the same semantics,
            # as-if `minimum_downgrade_revision` was specified, but we'll emit a warning
            # to suggest that feature is used instead.
            warnings.warn(NOT_IMPLEMENTED_WARNING.format(revision=revision))
            break

        except Exception as e:
            raise AlembicTestFailure(
                "Failed to downgrade through each revision individually.",
                context=[("Failing Revision", revision), ("Alembic Error", str(e))],
            )

    # We should only upgrade as far as we successfully downgraded.
    down_revisions = down_revisions[:index]

    for revision in reversed(down_revisions):
        try:
            alembic_runner.migrate_up_to(revision)
        except Exception as e:
            raise AlembicTestFailure(
                (
                    "Failed to upgrade through each revision individually after performing a "
                    "roundtrip upgrade -> downgrade -> upgrade cycle."
                ),
                context=[("Failing Revision", revision), ("Alembic Error", str(e))],
            )



import contextlib
import functools
from dataclasses import dataclass
from typing import Dict, List, Optional, Union

import alembic.command
import alembic.migration
from sqlalchemy.engine import Engine

from pytest_alembic.config import Config
from pytest_alembic.executor import CommandExecutor, ConnectionExecutor
from pytest_alembic.history import AlembicHistory
from pytest_alembic.revision_data import RevisionData


@contextlib.contextmanager
def runner(config: Config, engine=None):
    """Manage the alembic execution context, in a given context.

    Yields:
        `MigrationContext` to the caller.
    """
    command_executor = CommandExecutor.from_config(config)
    migration_context = MigrationContext.from_config(
        config, command_executor, ConnectionExecutor(), engine
    )

    command_executor.configure(connection=engine)
    yield migration_context


@dataclass
class MigrationContext:
    """Within a given environment/execution context, executes alembic commands."""

    command_executor: CommandExecutor
    revision_data: RevisionData
    connection_executor: ConnectionExecutor
    config: Config
    history: AlembicHistory
    connection: Engine = None

    @classmethod
    def from_config(
        cls,
        config: Config,
        command_executor: CommandExecutor,
        connection_executor: ConnectionExecutor,
        connection: Engine,
    ):
        raw_history = command_executor.script.revision_map
        history = AlembicHistory.parse(raw_history)

        return cls(
            command_executor=command_executor,
            revision_data=RevisionData.from_config(config),
            connection_executor=connection_executor,
            config=config,
            history=history,
            connection=connection,
        )

    @property
    def heads(self) -> List[str]:
        """Get the list of revision heads.

        Result is cached for the lifetime of the `MigrationContext`.
        """
        return self.command_executor.heads()

    @property
    def current(self) -> str:
        """Get the list of revision heads."""

        def get_current(conn):
            context = alembic.migration.MigrationContext.configure(conn)
            heads = context.get_current_heads()
            if heads:
                return heads[0]
            return None

        current = run_connection_task(self.connection, get_current)
        if current:
            return current
        return "base"

    def generate_revision(self, process_revision_directives=None, **kwargs):
        """Generate a test revision.

        The final act of this process raises a `RevisionSuccess`, which is used as a sentinal
        to indicate the revision was generated successfully, while not actually finishing the
        generation of the revision file.
        """
        alembic_config = self.command_executor.alembic_config
        config_directive = alembic_config.attributes["process_revision_directives"]
        fn = RevisionSuccess.process_revision_directives(
            _sequence_directives(config_directive, process_revision_directives)
        )
        try:
            return self.command_executor.run_command(
                "revision", process_revision_directives=fn, **kwargs
            )
        except RevisionSuccess:
            pass

    def raw_command(self, *args, **kwargs):
        """Execute a raw alembic command."""
        return self.command_executor.run_command(*args, **kwargs)

    def managed_upgrade(self, dest_revision):
        """Perform an upgrade, one migration at a time, inserting static data at the given points."""
        current = self.current
        for current_revision, next_revision in self.history.revision_window(current, dest_revision):
            before_upgrade_data = list(self.revision_data.get_before(next_revision))
            self.insert_into(data=before_upgrade_data, revision=current_revision, table=None)

            self.command_executor.upgrade(next_revision)

            at_upgrade_data = list(self.revision_data.get_at(next_revision))
            self.insert_into(data=at_upgrade_data, revision=next_revision, table=None)

        current = self.current
        return current

    def migrate_up_before(self, revision):
        """Migrate up to, but not including the given `revision`."""
        preceeding_revision = self.history.previous_revision(revision)
        return self.managed_upgrade(preceeding_revision)

    def migrate_up_to(self, revision):
        """Migrate up to, and including the given `revision`."""
        return self.managed_upgrade(revision)

    def migrate_up_one(self):
        """Migrate up by exactly one revision."""
        current = self.current
        next_revision = self.history.next_revision(current)
        new_revision = self.managed_upgrade(next_revision)
        if current == new_revision:
            return None
        return new_revision

    def migrate_down_before(self, revision):
        """Migrate down to, but not including the given `revision`."""
        next_revision = self.history.next_revision(revision)
        return self.migrate_down_to(next_revision)

    def migrate_down_to(self, revision):
        """Migrate down to, and including the given `revision`."""
        self.history.validate_revision(revision)
        self.command_executor.downgrade(revision)
        return revision

    def migrate_down_one(self):
        """Migrate down by exactly one revision."""
        previous_revision = self.history.previous_revision(self.current)
        self.command_executor.downgrade(previous_revision)
        return previous_revision

    def roundtrip_next_revision(self):
        """Upgrade, downgrade then upgrade.

        This is meant to ensure that the given revision is idempotent.
        """
        next_revision = self.migrate_up_one()
        if next_revision:
            self.migrate_down_one()
            return self.migrate_up_one()
        return None

    def insert_into(self, table: Optional[str], data: Union[Dict, List] = None, revision=None):
        """Insert data into a given table.

        Args:
            table: The name of the table to insert data into
            data: The data to insert. This is eventually passed through to SQLAlchemy's
                Table class `values` method, and so should accept either a list of
                `dict`s representing a list of rows, or a `dict` representing one row.
            revision: The revision of MetaData to use as the table definition for the insert.
        """
        if revision is None:
            revision = self.current

        return run_connection_task(
            self.connection,
            self.connection_executor.table_insert,
            revision=revision,
            tablename=table,
            data=data,
        )

    def table_at_revision(self, name, *, revision=None, schema=None):
        """Return a reference to a `sqlalchemy.Table` at the given revision.

        Args:
            name: The name of the table to produce a `sqlalchemy.Table` for.
            revision: The revision of the table to return.
            schema: The schema of the table.
        """
        revision = revision or self.current
        return self.connection_executor.table(
            self.connection, revision=revision, name=name, schema=schema
        )


class RevisionSuccess(Exception):
    """Raise when a revision is successfully generated.

    In order to prevent the generation of an actual revision file on disk when running tests,
    this exception should be raised.
    """

    @classmethod
    def process_revision_directives(cls, fn):
        """Wrap a real `process_revision_directives` function, while preventing it from completing."""

        @functools.wraps(fn)
        def _process_revision_directives(context, revision, directives):
            fn(context, revision, directives)
            raise cls()

        return _process_revision_directives


def _sequence_directives(*directives):
    def directive_wrapper(*args, **kwargs):
        for directive in directives:
            if not directive:
                continue
            directive(*args, **kwargs)

    return directive_wrapper


def run_connection_task(engine, fn, *args, **kwargs):
    """Run a given task on the provided connect, with the correct async/sync context.

    Given an async engine, we need to run the task in an async execution context,
    even though all internals are synchronous. This is how alembic suggests
    running the migrations themselves, so this matches that style.
    """
    # The user may not have sqlalchemy 1.4+, and therefore may not even be able to
    # use async engines.
    try:
        from sqlalchemy.ext.asyncio import AsyncEngine
    except ImportError:  # pragma: no cover
        AsyncEngine = None

    if AsyncEngine and isinstance(engine, AsyncEngine):
        import asyncio

        async def run(engine):
            async with engine.connect() as connection:
                result = await connection.run_sync(fn, *args, **kwargs)
                await connection.commit()

            await engine.dispose()
            return result

        return asyncio.run(run(engine))
    else:
        if isinstance(engine, Engine):
            with engine.connect() as connection:
                result = fn(connection, *args, **kwargs)
        else:
            result = fn(engine, *args, **kwargs)

        return result



from pytest_alembic.tests.default import (
    test_model_definitions_match_ddl,
    test_single_head_revision,
    test_up_down_consistency,
    test_upgrade,
)

__all__ = [
    "test_model_definitions_match_ddl",
    "test_single_head_revision",
    "test_up_down_consistency",
    "test_upgrade",
]



# flake8: noqa
from pytest_alembic.plugin.fixtures import alembic_config, alembic_engine, alembic_runner
from pytest_alembic.plugin.hooks import pytest_addoption, pytest_configure, pytest_sessionstart



import re
from dataclasses import dataclass
from pathlib import Path
from typing import Callable, Dict, List, Optional

import pytest
from _pytest import config

from pytest_alembic.plugin.error import AlembicReprError, AlembicTestFailure

pytest_version_tuple = getattr(pytest, "version_tuple", None)


@dataclass(eq=False)
class PytestAlembicPlugin:
    config: config.Config
    registered = False

    # Some weird decisions were made by pytest it seems like. There is not an obvious
    # way to support both <7 and >=7 without weird nonsense like this.
    if pytest_version_tuple and pytest_version_tuple[0] >= 7:

        def pytest_collect_file(self, file_path, path, parent):
            if self.should_register(file_path):
                return TestCollector.from_parent(parent, path=file_path)

    else:

        def pytest_collect_file(self, path, parent):  # type: ignore
            if self.should_register(Path(path)):
                return TestCollector.from_parent(parent, fspath=path)

    def should_register(self, path):
        if path.suffix != ".py":
            return False

        if not self.registered:
            self.registered = True
            return True

        return False

    def pytest_itemcollected(self, item):
        """Attach a marker to each test which uses the alembic fixture."""
        if not hasattr(item, "fixturenames"):
            return

        if "alembic_runner" in item.fixturenames:
            item.add_marker("alembic")


class TestCollector(pytest.Module):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._nodeid += "::pytest-alembic"
        self.add_marker("alembic")

    def collect(self):
        config = self.parent.config

        cli_enabled = config.option.pytest_alembic_enabled
        if not cli_enabled:
            return None

        option = config.option

        raw_included_tests = parse_test_names(config.getini("pytest_alembic_include"))
        raw_experimental_included_tests = parse_test_names(
            config.getini("pytest_alembic_include_experimental")
        )
        raw_excluded_tests = parse_test_names(
            option.pytest_alembic_exclude or config.getini("pytest_alembic_exclude")
        )

        test_collector = (
            TestOptionResolver.collect_test_definitions(default=True, experimental=True)
            .include(*raw_included_tests)
            .include_experimental(*raw_experimental_included_tests)
            .exclude(*raw_excluded_tests)
        )

        result = []
        for test in test_collector.sorted_tests():
            name = test.raw_name
            self.ihook.pytest_pycollect_makeitem(collector=self, name=name, obj=test)
            result.append(
                PytestAlembicItem.from_parent(
                    self,
                    name=name,
                    callobj=test.function,
                )
            )
        return result


class PytestAlembicItem(pytest.Function):
    def reportinfo(self):
        return (self.fspath, 0, f"[pytest-alembic] {self.name}")

    def repr_failure(self, excinfo):
        if isinstance(excinfo.value, AlembicTestFailure):
            return AlembicReprError(excinfo, self)
        return super().repr_failure(excinfo)


@dataclass(frozen=True)
class PytestAlembicTest:
    raw_name: str
    function: Callable
    is_experimental: bool

    @property
    def name(self):
        # Chop off the "test_" prefix.
        return self.raw_name[5:]


@dataclass
class TestOptionResolver:
    available_tests: Dict[str, PytestAlembicTest]

    included_tests: Optional[List[str]] = None
    included_experimental_tests: Optional[List[str]] = None
    excluded_tests: Optional[List[str]] = None

    @classmethod
    def collect_test_definitions(cls, default=True, experimental=True):
        import pytest_alembic.tests
        import pytest_alembic.tests.experimental

        test_groups = [(pytest_alembic.tests, False)]
        if experimental:
            test_groups.append((pytest_alembic.tests.experimental, True))

        all_tests = {}
        for test_group, is_experimental in test_groups:
            for name in dir(test_group):
                if name.startswith("test_"):
                    pytest_alembic_test = PytestAlembicTest(
                        name, getattr(test_group, name), is_experimental
                    )
                    all_tests[pytest_alembic_test.name] = pytest_alembic_test

        return cls(all_tests)

    def include(self, *tests):
        if tests:
            if self.included_tests is None:
                self.included_tests = []

            self.included_tests.extend(tests)
        return self

    def include_experimental(self, *tests):
        if tests:
            if self.included_experimental_tests is None:
                self.included_experimental_tests = []

            self.included_experimental_tests.extend(tests)
        return self

    def exclude(self, *tests):
        if tests:
            if self.excluded_tests is None:
                self.excluded_tests = []

            self.excluded_tests.extend(tests)
        return self

    def sorted_tests(self):
        return sorted(self.tests(), key=lambda t: t.raw_name)

    def tests(self):
        selected_tests = []
        invalid_tests = []

        excluded_set = set(self.excluded_tests or [])
        for excluded_test in excluded_set:
            if excluded_test not in self.available_tests:
                invalid_tests.append(excluded_test)

        if self.included_tests is None:
            included_tests = [
                t.name for t in self.available_tests.values() if t.is_experimental is False
            ]
        else:
            included_tests = self.included_tests

        for test_group in [included_tests, self.included_experimental_tests or []]:
            for included_test in test_group:
                if included_test in excluded_set:
                    continue

                if included_test not in self.available_tests:
                    invalid_tests.append(included_test)
                    continue

                selected_tests.append(included_test)

        if invalid_tests:
            invalid_str = ", ".join(sorted(invalid_tests))
            raise ValueError(f"The following tests were unrecognized: {invalid_str}")

        return [self.available_tests[t] for t in selected_tests]


def parse_test_names(raw_test_names):
    test_names = re.split(r"[,\n]", raw_test_names)

    result = set()
    for test_name in test_names:
        test_name = test_name.strip()
        if not test_name:
            continue
        result.add(test_name)
    return result



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['pytest_alembic',
 'pytest_alembic.plugin',
 'pytest_alembic.tests',
 'pytest_alembic.tests.experimental']

package_data = \
{'': ['*']}

install_requires = \
['alembic', 'pytest>=6.0', 'sqlalchemy']

entry_points = \
{'pytest11': ['pytest_alembic = pytest_alembic.plugin']}

setup_kwargs = {
    'name': 'pytest-alembic',
    'version': '0.9.0',
    'description': 'A pytest plugin for verifying alembic migrations.',
    'long_description': '![CircleCI](https://img.shields.io/circleci/build/gh/schireson/pytest-alembic/main)\n[![codecov](https://codecov.io/gh/schireson/pytest-alembic/branch/master/graph/badge.svg)](https://codecov.io/gh/schireson/pytest-alembic)\n[![Documentation Status](https://readthedocs.org/projects/pytest-alembic/badge/?version=latest)](https://pytest-alembic.readthedocs.io/en/latest/?badge=latest)\n\n## Introduction\n\nA pytest plugin to test alembic migrations (with default tests) and\nwhich enables you to write tests specific to your migrations.\n\n```bash\n$ pip install pytest-alembic\n$ pytest --test-alembic\n\n...\n::pytest_alembic/tests/model_definitions_match_ddl <- . PASSED           [ 25%]\n::pytest_alembic/tests/single_head_revision <- . PASSED                  [ 50%]\n::pytest_alembic/tests/up_down_consistency <- . PASSED                   [ 75%]\n::pytest_alembic/tests/upgrade <- . PASSED                               [100%]\n\n============================== 4 passed in 2.32s ===============================\n```\n\n## The pitch\n\nHave you ever merged a change to your models and you forgot to generate\na migration?\n\nHave you ever written a migration only to realize that it fails when\nthere’s data in the table?\n\nHave you ever written a **perfect** migration only to merge it and later\nfind out that someone else merged also merged a migration and your CD is\nnow broken!?\n\n`pytest-alembic` is meant to (with a little help) solve all these\nproblems and more. Note, due to a few different factors, there **may**\nbe some [minimal required\nsetup](http://pytest-alembic.readthedocs.io/en/latest/setup.html);\nhowever most of it is boilerplate akin to the setup required for alembic\nitself.\n\n### Built-in Tests\n\n- **test_single_head_revision**\n\n  Assert that there only exists one head revision.\n\n  We’re not sure what realistic scenario involves a diverging history to\n  be desirable. We have only seen it be the result of uncaught merge\n  conflicts resulting in a diverged history, which lazily breaks during\n  deployment.\n\n- **test_upgrade**\n\n  Assert that the revision history can be run through from base to head.\n\n- **test_model_definitions_match_ddl**\n\n  Assert that the state of the migrations matches the state of the\n  models describing the DDL.\n\n  In general, the set of migrations in the history should coalesce into\n  DDL which is described by the current set of models. Therefore, a call\n  to `revision --autogenerate` should always generate an empty migration\n  (e.g.\xa0find no difference between your database (i.e.\xa0migrations\n  history) and your models).\n\n- **test_up_down_consistency**\n\n  Assert that all downgrades succeed.\n\n  While downgrading may not be lossless operation data-wise, there’s a\n  theory of database migrations that says that the revisions in\n  existence for a database should be able to go from an entirely blank\n  schema to the finished product, and back again.\n\n- [Experimental\n  tests](http://pytest-alembic.readthedocs.io/en/latest/experimental_tests.html)\n\n  - all_models_register_on_metadata\n\n    Assert that all defined models are imported statically.\n\n    Prevents scenarios in which the minimal import of your models in your `env.py`\n    does not import all extant models, leading alembic to not autogenerate all\n    your models, or (worse!) suggest the deletion of tables which should still exist.\n\n  - downgrade_leaves_no_trace\n\n    Assert that there is no difference between the state of the database pre/post downgrade.\n\n    In essence this is a much more strict version of `test_up_down_consistency`,\n    where the state of a MetaData before and after a downgrade are identical as\n    far as alembic (autogenerate) is concerned.\n\n  These tests will need to be enabled manually because their semantics or API are\n  not yet guaranteed to stay the same. See the linked docs for more details!\n\nLet us know if you have any ideas for more built-in tests which would be\ngenerally useful for most alembic histories!\n\n### Custom Tests\n\nFor more information, see the docs for [custom\ntests](http://pytest-alembic.readthedocs.io/en/latest/custom_tests.html)\n(example below) or [custom static\ndata](http://pytest-alembic.readthedocs.io/en/latest/custom_data.html)\n(to be inserted automatically before a given revision).\n\nSometimes when writing a particularly gnarly data migration, it helps to\nbe able to practice a little timely TDD, since there’s always the\npotential you’ll trash your actual production data.\n\nWith `pytest-alembic`, you can write tests directly, in the same way\nthat you would normally, through the use of the `alembic_runner`\nfixture.\n\n```python\ndef test_gnarly_migration_xyz123(alembic_engine, alembic_runner):\n    # Migrate up to, but not including this new migration\n    alembic_runner.migrate_up_before(\'xyz123\')\n\n    # Perform some very specific data setup, because this migration is sooooo complex.\n    # ...\n    alembic_engine.execute(table.insert(id=1, name=\'foo\'))\n\n    alembic_runner.migrate_up_one()\n```\n\n`alembic_runner` has a number of methods designed to make it convenient\nto change the state of your database up, down, and all around.\n\n## Installing\n\n```bash\npip install "pytest-alembic"\n```\n',
    'author': 'Dan Cardin',
    'author_email': 'ddcardin@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/schireson/pytest-alembic',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4',
}


setup(**setup_kwargs)
